// Slowparse is a token stream parser for HTML and CSS text,
// recording regions of interest during the parse run and
// signaling any errors detected accompanied by relevant
// regions in the text stream, to make debugging easy. Each
// error type is documented in the [error specification][].
//
// Slowparse also builds a DOM as it goes, attaching metadata
// to each node build that points to where it came from in
// the original source.
//
// For more information on the rationale behind Slowparse, as
// well as its design goals, see the [README][].
//
// If [RequireJS] is detected, this file is defined as a module via
// `define()`. Otherwise, a global called `Slowparse` is exposed.
//
// ## Implementation
//
// Slowparse is effectively a finite state machine for
// HTML and CSS strings, and will switch between the HTML
// and CSS parsers while maintaining a single token stream.
//
//   [RequireJS]: http://requirejs.org/
//   [error specification]: spec/
//   [README]: https://github.com/mozilla/slowparse#readme
(function() {
  "use strict";

  var Stream = require("./Stream");
  var CSSParser = require("./CSSParser");
  var HTMLParser = require("./HTMLParser");
  var DOMBuilder = require("./DOMBuilder");

  // An internal function for performing a parse run
  // for either HTML or CSS content.
  function performParseRun(parser, domBuilder, errorDetectors, data) {
    data = data || "";

    var warnings = null,
        error = null;

    try {
      var _ = parser.parse();
      if (_.warnings) {
        warnings = _.warnings;
      }
    } catch (e) {
      if (e.parseInfo) {
        error = e.parseInfo;
      } else
        throw e;
    }

    errorDetectors.forEach(function(detector) {
      if (!error)
        error = detector(data, domBuilder.fragment.node) || null;
    });

    return {
      document: domBuilder.fragment.node,
      contexts: domBuilder.contexts,
      warnings: warnings,
      error: error
    };
  }

  // ### Exported Symbols
  //
  // `Slowparse` is the object that holds all exported symbols from
  // this library.
  var Slowparse = {
    // We export our list of recognized HTML elements and CSS properties
    // for clients to use if needed.
    HTML_ELEMENT_NAMES: HTMLParser.prototype.voidHtmlElements.concat(
                          HTMLParser.prototype.htmlElements.concat(
                            HTMLParser.prototype.obsoleteHtmlElements)),
    CSS_PROPERTY_NAMES: CSSParser.prototype.cssProperties,

    // We also export a few internal symbols for use by Slowparse's
    // testing suite.
    replaceEntityRefs: HTMLParser.replaceEntityRefs,
    Stream: Stream,

    // `Slowparse.HTML()` is the primary function we export. Given
    // a DOM document object (or a DOMBuilder instance) and a string
    // of HTML, we return an object with the following keys:
    //
    // * `document` is a DOM document fragment containing the DOM of
    //   the parsed HTML. If an error occurred while parsing, this
    //   document is incomplete, and represents what was built before
    //   the error was encountered.
    //
    // * `error` is a JSON-serializable object representing any error
    //   that occurred while parsing. If no errors occurred while parsing,
    //   its value is `null`. For a list of the types of errors that
    //   can be returned, see the [error specification][].
    //
    // An array of error detector functions can also be passed as a
    // third argument to this function. An error detector function takes
    // the HTML and generated document fragment as arguments and returns
    // an error object if an error is detected, or `undefined` otherwise.
    // This can be used for further error checking on the parsed document.
    //
    //   [error specification]: spec/
    HTML: function(document, html, options) {
      options = options || {};

      var stream = new Stream(html),
          disallowActiveAttributes = (typeof options.disallowActiveAttributes === "undefined") ? false : options.disallowActiveAttributes,
          domBuilder = new DOMBuilder(disallowActiveAttributes),
          parser = new HTMLParser(stream, domBuilder),
          errorDetectors = options.errorDetectors || [];

      return performParseRun(parser, domBuilder, errorDetectors, html);
    },

    // `Slowparse.CSS()` is a secondary function for CSS parsing. Given
    // a string of purse CSS, we return the same object as a call to
    // `Slowparse.HTML()`, with CSS-specific errors.
    //
    //   [error specification]: spec/
    CSS: function(css, options) {
      options = options || {};

      var stream = new Stream(css),
          domBuilder = new DOMBuilder(),
          parser = new CSSParser(stream, domBuilder),
          errorDetectors = options.errorDetectors || [];

      return performParseRun(parser, domBuilder, errorDetectors, css);
    },

    // `Slowparse.findError()` just returns any error in the given HTML
    // string, or `null` if the HTML contains no errors.
    findError: function(html, errorDetectors) {
      return this.HTML(document, html, errorDetectors).error;
    }
  };

  // AMD context
  if (typeof define !== "undefined") {
    define(function() { return Slowparse; });
  }

  // Node.js context
  else if(typeof module !== "undefined" && module.exports) {
    module.exports = Slowparse;
  }

  // browser context
  else if (typeof window !== "undefined") {
    window.Slowparse = Slowparse;
  }
}());
